/*	signal.c

	AFAPI signal definitions for SHMAPERS.

	July 1996 by H. Dietz
*/


#include "afapi.h"
#include "inline.h"


#define	S_SIGMAX	5
#define	S_EXIT		(S_SIGMAX - 1)


static int s_active = 0;	/* Signals enabled? */

/*	Pre-defined signal handlers...
*/

void
s_ignore(register p_bar m)
{
	/* Nothing to do...  that's why it's called s_ignore() */
}


static void
s_exitHandler(register p_bar m)
{
	/* Is every process making this request? */
	if (m == (((p_bar) (1 << NPROC)) - ((p_bar)1))) {
		/* Yup, everybody wants out...  do it */
		s_active = 0;
	}
}

/*	The RSM handler...

	Really pretty simple stuff.  A pre-installed signal handler is
	defined (s_updateHandler()) as a static function.  Everything
	else is just simple interface crud....
*/

struct s_updateArgsType s_updateArgs;	/* Where the args go... */


static void
s_updateHandler(register p_bar m)
{
	register int i;

	/* For each process that wants to write... */
	for (i=0; i<NPROC; ++i) {
		if (m & ((p_bar)(1 << i))) {
			if (i == IPROC) {
				/* Send object length */
				p_bcastPut32(s_updateArgs.n);

				if (s_updateArgs.n > 0) {
					/* Send the address, then data */
					p_bcastPutptr(s_updateArgs.a);
					p_bcastPutz(s_updateArgs.a,
						    s_updateArgs.n);
				}
			} else {
				/* Anything to get? */
				register int n = p_bcastGet32();

				if (n > 0) {
					register void *p = p_bcastGetptr();

					p_bcastGetz(p, n);

					/* Was our write clobbered? */
					if ((p <= s_updateArgs.a) &&
					    ((p + n) >=
					     (s_updateArgs.a +
					      s_updateArgs.n))) {
						s_updateArgs.n = 0;
					}
				}
			}
		}
	}
}

/*	s_vector[] array...

	The following is the magic internal structure used to vector
	signals to the correct handlers.  It is specific to SHMAPERS.
	The entry for s_updateHandler must be in the S_SIGMEM position
	in the table, just as s_exitHandler must be in the S_SIGMAX-1
	position.
*/

static void (*s_vector[S_SIGMAX])(register p_bar mask) = {
	/* Normal signals... */
	s_updateHandler,
	s_ignore,
	s_ignore,
	s_ignore,

	/* And one signal you can't redefine... */
	s_exitHandler
};

/*	s_install()

	Install a signal handler...
*/

int
s_install(register int s,
register void (*fn)(p_bar m))
{
	/* Install the signal handler fn()... */
	register int which = s;

	/* If s is S_SIGANY, pick one that is s_ignore()... */
	if (which == S_SIGANY) {
		for (which = 0; which<(S_SIGMAX-1); ++which) {
			if (s_vector[which] == s_ignore) break;
		}
	}

	/* Got a valid signal number? */
	if ((which < 0) || (which >= (S_SIGMAX-1))) {
		/* Nope.  Return with error. */
		return(-1);
	}

	/* Ok, install it... */
	s_vector[which] = fn;

	/* Return signal number */
	return(which);
}

/*	s_init() and s_exit()...
*/

int
s_init(void)
{
	/* If it's already active, oops... */
	if (s_active) return(1);

	/* Ok, make it active... */
	s_active = 1;
	return(0);
}


void
s_exit(void)
{
	/* Shut-down the signal system using a signal... */
	while (s_active) {
		s_signal(S_EXIT);
	}

	/* Then make sure that everybody got out... */
	p_wait();
}

/*	s_signal()...

	This code must recognize who has sent signals, create masks
	grouping similar signals together, and invoke the appropriate
	handler functions in a way that gives RSM the highest priority
	and shared memory exit the lowest priority....
*/

void
s_signal(register int s)
{
	p_32 sigs[NPROC];
	register int sig;

	/* Send our signal and collect results... */
	p_gather32(&(sigs[0]), s);

	/* Scan for signals in priority order... */
	for (sig=0; sig<S_SIGMAX; ++sig) {
		register p_bar mask = 0;
		register p_bar bit = 1;
		register int i;

		/* Collect similar signals... */
		for (i=0; i<NPROC; ++i) {
			if (sigs[i] == sig) {
				mask |= bit;
			}

			bit <<= 1;
		}

		/* If we found somebody, call the handler... */
		if (mask != 0) {
			s_vector[sig](mask);
		}
	}
}

/*	Inlined signal and RSM stuff...

	These are just callable versions of the simple interface
	routines....
*/

void
s_poll(void)
{
	s__poll();
}


void
s_wait(void)
{
	s__wait();
}


void
s_update(register void *a,
register int n)
{
	s__update(a, n);
}


void
s_initLock(s_lock *lock)
{
	s_initLock(lock);
}


void
s_acquire(s_lock *lock)
{
	s__acquire(lock);
}


void
s_release(s_lock *lock)
{
	s__release(lock);
}
